bl_info = {
    "name": "Favorite Objects Manager",
    "author": "1P2D",
    "version": (1, 0, 3),
    "blender": (4, 3, 2),
    "description": "Manage favorite objects with custom shortcut and pie menu",
    "category": "Object",
}

import bpy
from bpy.types import (
    Operator,
    Menu,
    PropertyGroup,
)
from bpy.props import (
    StringProperty,
    BoolProperty,
)

import bpy, weakref
from bpy.app.handlers import persistent

# Track existing objects using a weak set
known_objects = weakref.WeakSet()

@persistent
def reset_favorite_on_new_objects(scene):
    global known_objects
    
    # Initialize known_objects if it's empty (first run)
    if not known_objects:
        for s in bpy.data.scenes:
            known_objects.update(s.objects)
    
    current_objs = set(scene.objects)
    new_objs = current_objs - known_objects
    for obj in new_objs:
        # Check if this object is newly created (not loaded from file)
        if hasattr(obj, "favorite_object"):
            obj.favorite_object.is_favorite = False
    known_objects.update(current_objs)  # Use update() for WeakSet

@persistent
def handle_object_rename(scene):
    # This handler will be called on each depsgraph update
    # We need to check if any objects have been renamed
    for obj in scene.objects:
        # If this object has a pointer with a different name in favorite_objects,
        # we need to update the favorite_objects entry
        if obj.name in scene.objects and obj.name not in scene.favorite_objects:
            # Check if this might be a renamed object
            for old_name, item in scene.favorite_objects.items():
                if old_name not in scene.objects and item.is_favorite:
                    # This could be a renamed object - let's update the entry
                    item.name = obj.name
                    if hasattr(item, "original_name"):
                        del item["original_name"]
                    break

# Custom property to mark objects as favorites
class FavoriteObjectProperties(PropertyGroup):
    is_favorite: BoolProperty(
        name="Is Favorite",
        default=False,
        description="Mark object as favorite in the current scene",
        options={'LIBRARY_EDITABLE'}  # Make the property persistent
    )

# Add scene property to store favorite objects
class FavoriteObjectsCollection(PropertyGroup):
    is_favorite: BoolProperty(
        name="Is Favorite",
        default=False,
        description="Mark object as favorite in this scene",
    )
    original_name: StringProperty(
        name="Original Name",
        default="",
        description="Original name of the object when it was marked as favorite",
    )

# Pie menu operator
class VIEW3D_MT_PIE_favorite_objects(Menu):
    bl_label = "Favorite Objects"

    @classmethod
    def poll(cls, context):
        if context.mode != 'OBJECT':
            return False
        return True

    def draw(self, context):
        layout = self.layout
        scene = context.scene
        
        # Get all favorite objects for current scene
        favorite_objects = [obj for obj in scene.objects 
                          if obj.name in scene.favorite_objects and 
                          scene.favorite_objects[obj.name].is_favorite]
        
        # Limit to 8 objects for pie menu
        favorite_objects = favorite_objects[:8]
        
        # Update menu label based on favorites
        if not favorite_objects:
            self.bl_label = "Favorite Objects (empty)"
        else:
            self.bl_label = "Favorite Objects"
            
        pie = layout.menu_pie()
        for i, obj in enumerate(favorite_objects):
            pie.operator("object.select_favorite_object", text=obj.name).object_name = obj.name

# Operator to select favorite object
class SelectFavoriteObject(Operator):
    bl_idname = "object.select_favorite_object"
    bl_label = "Select Favorite Object"
    bl_options = {'REGISTER', 'UNDO'}
    
    object_name: StringProperty()
    
    def execute(self, context):
        # Deselect all objects
        bpy.ops.object.select_all(action='DESELECT')
        
        # Get the favorite object
        obj = bpy.data.objects.get(self.object_name)
        if not obj:
            self.report({'WARNING'}, f"Object '{self.object_name}' not found")
            return {'CANCELLED'}

        # Check if object is in an excluded collection
        if obj.name not in context.view_layer.objects:
            self.report({'WARNING'}, f"'{obj.name}' is in an excluded collection")
            return {'CANCELLED'}
            
        # Check if object is hidden in viewport
        if obj.hide_viewport:
            self.report({'WARNING'}, f"'{obj.name}' is hidden in viewport")
            return {'CANCELLED'}
            
        # Check if object is hidden in view layer
        if obj.hide_get():
            self.report({'WARNING'}, f"'{obj.name}' is hidden in view layer")
            return {'CANCELLED'}
            
        # Check if object has selection disabled
        if not obj.hide_select and obj.visible_get():
            obj.select_set(True)
            context.view_layer.objects.active = obj
        else:
            self.report({'WARNING'}, f"'{obj.name}' has selection disabled or is not visible")
            return {'CANCELLED'}
        
        return {'FINISHED'}

    @classmethod
    def poll(cls, context):
        if context.mode != 'OBJECT':
            cls.poll_message_set("Only available in Object Mode")
            return False
        return True

# Toggle favorite status operator
class ToggleFavoriteObject(Operator):
    bl_idname = "object.toggle_favorite"
    bl_label = "Toggle Favorite Object"
    bl_options = {'REGISTER', 'UNDO'}
    
    @classmethod
    def description(cls, context, properties):
        obj = context.active_object
        if obj and obj.name in context.scene.favorite_objects and context.scene.favorite_objects[obj.name].is_favorite:
            return "Remove selected object(s) from your favorites list"
        return "Add selected object(s) to your favorites list for quick access"
    
    def execute(self, context):
        active_obj = context.active_object
        if not active_obj:
            return {'CANCELLED'}
            
        scene = context.scene
        
        # Get the target state based on active object (inverse of current state)
        current_state = scene.favorite_objects.get(active_obj.name, {}).get("is_favorite", False)
        target_state = not current_state
        
        if target_state:  # If we're trying to add to favorites
            # Count current favorites
            current_favorites = len([obj for obj in scene.objects 
                                  if obj.name in scene.favorite_objects and 
                                  scene.favorite_objects[obj.name].is_favorite])
            # Calculate how many new objects we're trying to add
            new_objects = len([obj for obj in context.selected_objects 
                             if obj.name not in scene.favorite_objects or
                             not scene.favorite_objects[obj.name].is_favorite])
            
            # Check if adding these would exceed the limit
            if current_favorites + new_objects > 8:
                self.report({'WARNING'}, "Cannot add more than 8 favorite objects (Pie menu limitation)")
                return {'CANCELLED'}
        
        # Apply to all selected objects
        modified_count = 0
        for obj in context.selected_objects:
            if obj.name not in scene.favorite_objects:
                item = scene.favorite_objects.add()
                item.name = obj.name
                item.original_name = obj.name
            scene.favorite_objects[obj.name].is_favorite = target_state
            modified_count += 1
            
        # Add status message
        status = "removed from" if not target_state else "added to"
        if modified_count > 1:
            self.report({'INFO'}, f"{modified_count} objects {status} favorites")
        else:
            self.report({'INFO'}, f"'{active_obj.name}' {status} favorites")
            
        return {'FINISHED'}
    
    @classmethod
    def poll(cls, context):
        if context.mode != 'OBJECT':
            cls.poll_message_set("Only available in Object Mode")
            return False
        return context.active_object is not None

# Add to menu
def menu_func(self, context):
    if context.mode == 'OBJECT':  # Only show menu item in Object Mode
        obj = context.active_object
        if obj:
            scene = context.scene
            is_favorite = scene.favorite_objects.get(obj.name, {}).get("is_favorite", False)
            text = "Remove from Favorite Objects" if is_favorite else "Add to Favorite Objects"
            self.layout.operator(ToggleFavoriteObject.bl_idname, text=text)

# Import the new preferences module
from . import preferences

# Keymap
addon_keymaps = []

def register():
    # Register classes
    bpy.utils.register_class(FavoriteObjectProperties)
    bpy.utils.register_class(FavoriteObjectsCollection)
    bpy.utils.register_class(VIEW3D_MT_PIE_favorite_objects)
    bpy.utils.register_class(SelectFavoriteObject)
    bpy.utils.register_class(ToggleFavoriteObject)
    preferences.register()
    
    global known_objects
    known_objects = weakref.WeakSet()  # Initialize as empty WeakSet
    
    # Register handlers
    bpy.app.handlers.depsgraph_update_post.append(reset_favorite_on_new_objects)
    bpy.app.handlers.depsgraph_update_post.append(handle_object_rename)
    
    # Register properties
    bpy.types.Object.favorite_object = bpy.props.PointerProperty(type=FavoriteObjectProperties)
    bpy.types.Scene.favorite_objects = bpy.props.CollectionProperty(type=FavoriteObjectsCollection)
    
    # Add to menu
    bpy.types.VIEW3D_MT_object_context_menu.append(menu_func)
    
    # Add keymap
    wm = bpy.context.window_manager
    kc = wm.keyconfigs.addon
    if kc:
        km = kc.keymaps.new(name='3D View', space_type='VIEW_3D')
        kmi = km.keymap_items.new('wm.call_menu_pie', 'LEFTMOUSE', 'PRESS', 
                                 ctrl=True, alt=True)
        kmi.properties.name = "VIEW3D_MT_PIE_favorite_objects"
        addon_keymaps.append((km, kmi))

def unregister():
    # Remove keymap
    for km, kmi in addon_keymaps:
        km.keymap_items.remove(kmi)
    addon_keymaps.clear()
    
    # Remove from menu
    bpy.types.VIEW3D_MT_object_context_menu.remove(menu_func)

    # Remove handlers
    bpy.app.handlers.depsgraph_update_post.remove(reset_favorite_on_new_objects)
    bpy.app.handlers.depsgraph_update_post.remove(handle_object_rename)
    
    # Remove scene property
    del bpy.types.Scene.favorite_objects
    
    # Unregister classes
    preferences.unregister()
    bpy.utils.unregister_class(ToggleFavoriteObject)
    bpy.utils.unregister_class(SelectFavoriteObject)
    bpy.utils.unregister_class(VIEW3D_MT_PIE_favorite_objects)
    bpy.utils.unregister_class(FavoriteObjectsCollection)
    bpy.utils.unregister_class(FavoriteObjectProperties)

if __name__ == "__main__":
    register()